/*********************************************/
/*         IRS Code detect Device Driver Program         */
/*                                           */
/*                  developed by Shinji Ushimaru */
/*                     ver 1.0  Mar. 16 2004 */
/*********************************************/

#define __KERNEL__
#define MODULE

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/ioctl.h>
#include <linux/sched.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/uaccess.h>

#include "irs_code.h"

static unsigned int busy = 0;

static int irs_open(struct inode *inode, struct file *filp){
  if (busy) return - EBUSY;
  
  //printk("irs_open:\n");
  //printk("   filp->f_version     : %lu \n",filp->f_version);

  linecount = LINECOUNT;

  //printk("   LINECOUNT: %d \n",linecount);

  busy = 1;
  MOD_INC_USE_COUNT;

  //irs_port_region(IRS_PORT_SEL,IRS_PORT_DET);
  irs_port_region(IRS_PORT_SEL,4);

  return 0;
}

static int irs_release(struct inode *inode, struct file *filp){
  busy = 0;
  MOD_DEC_USE_COUNT;

  //printk("irs_close:\n");
  //printk("   filp->f_version     : %lu \n",filp->f_version);

  //release_region(IRS_PORT_SEL,IRS_PORT_DET);
  release_region(IRS_PORT_SEL,4);
  //printk("irs_release:\n");

  return 0;
}

static ssize_t irs_sel(struct file *filp, const char *buf, size_t count, loff_t *offset){
  int len;
  int *selno;
  static int msg[10]={};

  selno=msg;

  if (count < 0){
     return -EINVAL;
  }

     len = count;

     copy_from_user(selno,buf,len);
     //printk("sel number = %d \n",*selno);

     outw(*selno,IRS_PORT_SEL);

  return len;
}

static ssize_t irs_det(struct file *file, char *buf, size_t count, loff_t *offset){
static char dummy[1]={};
  char *dat;
  int data,data0;
  dat=dummy;

  if (linecount == 0){
     return 0;
  }else{
     linecount--;
  }

     data0 = inw(IRS_PORT_DET);
     data = data0 & 0x8000;
	if(data == 0) *dat='0';
	else{
		data = data0 & 0x000f;
		if(data==0x01) *dat='1';
		if(data==0x02) *dat='2';
		if(data==0x03) *dat='3';
		if(data==0x04) *dat='4';
		if(data==0x05) *dat='5';
		if(data==0x06) *dat='6';
	}

	//printk("data =%x\n",data);
	copy_to_user(buf,dat,1);

	return(0);
}

static struct file_operations irs_fops = {
   owner:              THIS_MODULE,
   read:               irs_det,
   write:              irs_sel,
   ioctl:              NULL,
   open:               irs_open,
   release:            irs_release,
 };


static int irs_port_region(unsigned int port, unsigned int range){
  int err;

  //printk("irs_port_region:\n");

  if ((err=check_region(port,range)) < 0 ){
    //printk("   This ports <%x - %x> are using now!\n",port,range);
    //printk("   check_region(port,range) : %d\n",err);
    /* return err; */
  }     /* using port now */
  
  request_region(port,range,"irs");     /* always success */

  //printk(" I/Oport from: %x   to: %x \n",port,range);

  return 0;

}

int init_module (void){
  int i;

  //printk("install 'irs' into major %d\n",IRS_MAJOR);
  i = register_chrdev (IRS_MAJOR, "irs", &irs_fops);

  if (i){
     //printk("device registration error\n");
     return -EBUSY;
  }

  return 0;
} 

void cleanup_module (void){
  int i;

  //printk("remove 'irs' from major %d\n",IRS_MAJOR);
  i = unregister_chrdev (IRS_MAJOR,"irs");

  if (i){
     //printk ("unregister_chrdev failed\n");
  }

}
