/*************************************************************
*  pwm_enc_module2.c  pwm & encoder RT-Task Module  Ver.1.0  *
*                                          Date:14 May 2004  *
*                                Programed by Hiroki.Segawa  *
*************************************************************/

/*******************************************************************
*       This Source file contains these function.                  *
*                                                                  *
* void *thread_code(void *t);           RT-Task Module Main thread *
* int my_handler(unsigned int fifo);    Command handler            *
* int init_module(void);                Module initialize method   *
* void cleanup_module(void);            Clean Up method            *
*******************************************************************/

#include <linux/errno.h>
#include <asm/io.h>
#include <unistd.h>
#include <math.h>
#include <rtl.h>
#include <time.h>
#include <rtl_sched.h>
#include <rtl_fifo.h>
#include "pwm_enc.h"
#include "pwm_module.h"

static pthread_t tasks[1];
static int fflag = 0;
static float dt;

void *thread_code(void *t){
	// variables
	int fifo3=3;
	int fifo4=fifo3+1;
	struct pwm_speed_data pwm;
	struct count_data count;
	
	int pwm_data_l1=0,pwm_data_r1=0;
	int pwm_data_l2=0,pwm_data_r2=0;
	float ua[2]={0.,0.};
	float xa_tmp[6];
	static float xa[6]={0.,0.,0.,0.,0.,0.};
	static float ya[2]={0.,0.};
	static int num;
	static float ref[2];
	float e[2],esum[2]={0.,0.};
	float eta[4];
	int dl=0,dr=0,i,j;
	int v=0,x=0,y,w;
	int abs1,abs2;
	int ret;
	int err;
	int Tup=0;
	static int counter=0;
	
	while(1){
		y = 0; w = 0;
		ret = pthread_wait_np(); // wait next period
		counter++;
		// Get pwm data from FIFO
		while((err = rtf_get(fifo4, &pwm, sizeof(pwm))) == sizeof(pwm)){
			pwm_data_l1 = pwm.l_speed;
			pwm_data_r1 = pwm.r_speed;
  			pwm_data_l2 = 0;
			pwm_data_r2 = 0;
			Tup = pwm.tup;
		}
		// step calculation range(step)
		num = Tup/dt;
		// calculation of target value
		if(counter*dt<=Tup){
			ref[0] = pwm_data_r1 * (2. / (diameter)) * counter * dt / Tup;
			ref[1] = pwm_data_l1 * (2. / (diameter)) * counter * dt / Tup;
		}
		eta[0] = ref[0];
		eta[1] = ref[1];
		eta[2] = ya[0];
		eta[3] = ya[1];
		
		switch(pwm.mode){
			case DUTY_MODE:
				// Data Write in I/O Port
				outw(pwm_data_l1, PWM_LEFT_WRITE);
				outw(pwm_data_r1, PWM_RIGHT_WRITE);
				// initialization of variables
				for(i=0;i<2;i++){
					ua[i] = 0.;
					ya[i] = 0.;
					ref[i] = 0.;
					e[i] = 0.;
					esum[i] = 0.;
				}
				for(i=0;i<6;i++){
					xa[i] = 0.;
					xa_tmp[i] = 0.;
				}
				for(i=0;i<4;i++){
					eta[i] = 0.;
				}
				Tup = 0; counter = 0;
				count.right_count = 0; count.left_count = 0;
				count.left_sum = 0; count.right_sum = 0;
				dl=0; dr=0;
				v=0; x=0;
				abs1=0; abs2=0;
				pwm_data_l2 = 0; pwm_data_r2 = 0;
				break;
				
			case PID_MODE:
				// PID Control
				v = pwm_data_l1 + pwm_data_l2;
				if(v>Vmin){
					w = Kpl * v + vshift;
				}else if(v<-Vmin){
					w = Kpl * v - vshift;
				}
				w = (w>95) ? 95 : w;
				w = (w<-95) ? -95 : w;
				w = (w>=0) ? w : (-w+128);
				x = pwm_data_r1 - pwm_data_r2;
				// Shift of Duty
				if(x>Vmin){
					y = Kpr * x + vshift;
				}else if(x<-Vmin){
					y = Kpr * x - vshift;
				}
				// if y>limit
				y = (y>95) ? 95 : y;
				y = (y<-95) ? -95 : y;
				y = (y>=0) ? y : (-y+128);
				abs1 = (v>=0) ? v : -v;
				abs2 = (x>=0) ? x : -x;
				count.br = 0;
				// if control end, count.br=1
				if(abs1<=Vmin && abs2<=Vmin){
					count.br = 1;
				}else{
					count.br = 0;
				}
				// Data Write in PWM I/O Port
				outw(w, PWM_LEFT_WRITE);
				outw(y, PWM_RIGHT_WRITE);
				break;
				
			case LQR_MODE:
				//LQR Control
				//calculation of error
				for(i=0;i<2;i++){
					e[i] = ref[i] - ya[i];
					esum[i] += e[i];
				}
				for(i=0;i<2;i++){
					ua[i] = 0.0;
				}
				// calculation of output u
				for(i=0;i<2;i++){
					for(j=0;j<6;j++){
						ua[i] += Cd[i][j] * xa[j];
					}
				}
				// if ua>limit
				for(i=0;i<2;i++){
					if(ua[i]>Vmax){
						ua[i] = Vmax;
					}
					if(ua[i]<-Vmax){
						ua[i] = -Vmax;
					}
				}
				// Convert to Duty from output
				dr = 127 / Vmax * ua[0];
				dl = 127 / Vmax * ua[1];
				// Shift of Duty
				if(dr>=0){
					dr = dr + drshift;
					if(dr>=dmax){
						dr = dmax;
					}
				}
				if(dr<0){
					dr = dr - drshift;
					if(dr<=-dmax){
						dr = -dmax;
					}
				}
				if(dl>=0){
					dl = dl + dlshift;
					if(dl>=dmax){
						dl = dmax;
					}
				}
				if(dl<0){
					dl = dl - dlshift;
					if(dl<-dmax){
						dl = -dmax;
					}
				}
				dr = (dr>=0) ? dr : (-dr+128);
				dl = (dl>=0) ? dl : (-dl+128);
				
				abs1 = (e[0]>=0) ? e[0] : -e[0];
				abs2 = (e[1]>=0) ? e[1] : -e[1];
				count.br = 0;
				// if e<error max, control end(count.br=1)
				if(counter>num){
					if(abs1<=errmax && abs2<=errmax){
						count.br = 1;
					}else{
						count.br = 0;
					}
				}
				if(counter>=maxcount){
					count.br=1;
				}
				// Data Write in PWM I/O Port
				outw(dl, PWM_LEFT_WRITE);
				outw(dr, PWM_RIGHT_WRITE);
				break;

			default:
				break;
		}
		// Read Data from Encoder I/O Port
		count.left_count = inw(LEFT_COUNTER);
		outw(COUNTER_RESET, LEFT_COUNTER);
		count.right_count = inw(RIGHT_COUNTER);
		outw(COUNTER_RESET, RIGHT_COUNTER);
		// if count is minus,convert
		if(((count.left_count >> 15) & 1) == 1)
			count.left_count = -1 * (0x10000 - count.left_count);
 		pwm_data_l2 += count.left_count;  
		if(((count.right_count >> 15) & 1) == 1)
			count.right_count = -1 * (0x10000 - count.right_count);
		pwm_data_r2 += count.right_count;
		// sum up of encoder count
		count.left_sum = pwm_data_l2;
		count.right_sum = pwm_data_r2;
		
		// calculation of rotated angle[rad]
		ya[0] = (2. * pi) / 1600. * (float)(pwm_data_r2);
		ya[1] = -(2. * pi) / 1600. * (float)(pwm_data_l2);
		// calculation of matrix
		for(i=0;i<6;i++){
			xa_tmp[i] = 0.;
			for(j=0;j<6;j++)
				xa_tmp[i] += Ad[i][j] * xa[j];
			for(j=0;j<4;j++)
				xa_tmp[i] += Bd[i][j] * eta[j];
		}
		for(i=0;i<6;i++){
			xa[i] = xa_tmp[i];
		}
		// Encoder Count Write in FIFO
		rtf_put(fifo3, &count, sizeof(count));
	}
	return 0;
}

// Command Handler
int my_handler(unsigned int fifo){
	struct my_msg_struct msg;
	int err;

	// Get Command Data from FIFO
	while((err = rtf_get(COMMAND_FIFO, &msg, sizeof(msg))) == sizeof(msg)){
		rtf_put(msg.task + TASK_CONTROL_FIFO_OFFSET, &msg, sizeof(msg));
		dt = msg.period; /* [msec] */
			switch(msg.command){
				case START_TASK:
					outw(COUNTER_RESET, LEFT_COUNTER);
					outw(COUNTER_RESET, RIGHT_COUNTER);
					// Apply Period of thread
					pthread_make_periodic_np(tasks[0], gethrtime(), (long long)msg.period * (long long)1000000);
					// thread wakeup
					pthread_wakeup_np(tasks[0]);
					break;
				case STOP_TASK:	
					outw(0, PWM_LEFT_WRITE);
					outw(0, PWM_RIGHT_WRITE);
					// thread suspend
					pthread_suspend_np(tasks[0]);
					break;
				case RESTART_TASK:
					outw(COUNTER_RESET, LEFT_COUNTER);
					outw(COUNTER_RESET, RIGHT_COUNTER);
					// thread wakeup
					pthread_wakeup_np(tasks[0]);
					break;
				default:
					return 0;
			}
	}
	if(err!=0){
		return -EINVAL;
	}
	return 0;
}

// initialization
int init_module(void){
	pthread_attr_t attr;
	struct sched_param sched_param;
	int c[4];
	int ret;

	rtf_destroy(3);
	rtf_destroy(4);
	rtf_destroy(5);
	rtf_destroy(6);
	c[0] = rtf_create(3, 4000);
	c[1] = rtf_create(4, 4000);
	c[2] = rtf_create(5, 200);
	c[3] = rtf_create(6, 100);
	rtl_printf("Fifo return 1=%d 2=%d 3=%d\n",c[0],c[1],c[2]);

	pthread_attr_init(&attr);
	sched_param.sched_priority = 5;
	pthread_attr_setschedparam(&attr, &sched_param);
	ret = pthread_create(&tasks[0], &attr, thread_code, (void *)2);
	pthread_setfp_np (tasks[0],1);
	fflag = 1;

	rtf_create_handler(5, &my_handler);
	return 0;
}

// Destroy FIFO
void cleanup_module(void){
	outw(0, PWM_LEFT_WRITE);
	outw(0, PWM_RIGHT_WRITE);
#ifdef DEBUG
	rtl_printf("%d\n", rtf_destroy(3));
	rtl_printf("%d\n", rtf_destroy(4));
	rtl_printf("%d\n", rtf_destroy(5));
	rtl_printf("%d\n", rtf_destroy(6));
#else
	rtf_destroy(3);
	rtf_destroy(4); 
	rtf_destroy(5);
	rtf_destroy(6);
#endif
	if(fflag==1){
		pthread_delete_np (tasks[0]);
	}
}
